//  analyse_template()
//  analyse_processor()
//
//  add_processor_declaration()
//  get_template_local_size()
//  add_processor_declaration()

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include "compilerApi.h"

#include "anls_att.h"

void analyse_template(SgStatement *);
void show_template(SgStatement *);
void add_template_declararion(SgStatement *);

void fill_template_table(Template_def * tdef, SgExpression * expr)
{
  //printf("Current Expr: %d\n", expr->id());
  switch(expr->variant()) {
  case EXPR_LIST:
    fill_template_table(tdef, expr->lhs() );
    if (expr->operand(2))
      fill_template_table(tdef, expr->operand(2));
    break;

  case DDOT:
	// if expr->lhs() is NULL , it must be assumed size array etc .
    if (expr->lhs() == NULL)
	tdef->add_dimention(NULL, NULL);
    else
	fill_template_table(tdef, expr->lhs());
    break;

  case RANGE_OP:
    {
      SgExpression *start, *end;

      start = expr->operand(1);
      if (start) {
       if (start->isInteger())
      	start = calculateExp(start);
       else 
	if (!silentFlag) printf("Warning:Unknow template boundaries.\n");
      }
      else
	start=new SgValueExp(1);

      end = expr->operand(2);
      if (end->isInteger())
      	end = calculateExp(end);
      else
	if (!silentFlag) printf("Warning:Unknow template boundaries.\n");

      tdef->add_dimention(start, end);
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
  };
}

void deal_template_array(SgStatement * s, SgExpression * expr)
{
  switch(expr->variant()) {
  case EXPR_LIST:
    deal_template_array(s, expr->operand(1));
    if (expr->operand(2))
      deal_template_array(s, expr->operand(2));
    break;
    
  case ARRAY_REF:
    {
      Template_def * tdef =new Template_def;
      tdef->setsymbol(expr->symbol());
      fill_template_table(tdef, expr->operand(1));
      s->controlParent()->addAttribute(Temp_Def, (void *) tdef, sizeof(Template_def));
      expr->symbol()->addAttribute(Temp_Def, (void *) tdef, sizeof(Template_def));
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
    break;

  };
}

void analyse_template(SgStatement * s)
{
  deal_template_array(s, s->expr(0));
}

void show_template(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    ((Template_def *) s->attributeValue(i, Temp_Def)) ->show();
  };
}

void add_template_declaration(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    Template_def * tdef = (Template_def *)s->attributeValue(i, Temp_Def);

    SgSymbol * tempname = s->getSymbolInCurrentScope(mixstring, "rng", tdef->getsymbol()->identifier());

    tdef->setrng(tempname);

    SgValueExp * dim = new SgValueExp (tdef->getDim());
    
    SgExpression * range = new SgExpression(RANGE_OP, NULL, dim, NULL);

    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);

    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

    SgExpression * array = new SgExpression(ARRAY_REF, index, NULL, tempname);

    SgType * stype = SgTypeInt();

    SgExpression * arraytype = new SgExpression(TYPE_OP, NULL, NULL, NULL, stype);

    SgStatement * decltemplate =new SgStatement(VAR_DECL, NULL, NULL, array, arraytype, NULL);
  
    s->childList1(0)->insertStmtBefore(*decltemplate, *s);

  };
}

//-----------------------;

void analyse_processor(SgStatement *);
void show_processor(SgStatement *);
void add_processor_declaration(SgStatement *);

void fill_table(Processor_def * pdef,SgExpression * expr)
{
  //printf("Current Expr: %d\n", expr->id());
  switch(expr->variant()) {
  case EXPR_LIST:
    fill_table(pdef, expr->operand(1));
    if (expr->operand(2))
      fill_table(pdef, expr->operand(2));
    break;

  case DDOT:
    fill_table(pdef, expr->operand(1));
    break;

  case RANGE_OP:
    {
      SgExpression *start, *end;
      if (expr->operand(1)) 
	  start=expr->operand(1);
      else
	start=new SgValueExp(1);

      end=expr->operand(2);
 
      pdef->add_dimention(start, end);
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
  };
}

void intrinsic_functions_in_size(SgStatement *s, SgExpression *size) 
{
        if (size)
        switch (size->variant()) {
        case INTRIN_FUNC_CALL:
//                replace_intrinsic_function_name(size,0);
      		analyse_expression(s,size);
                break;
        default:
                if (size->lhs()) intrinsic_functions_in_size(s,size->lhs());
                if (size->rhs()) intrinsic_functions_in_size(s,size->rhs());
        }
}

void deal_processor_array(SgStatement * s, SgExpression * expr)
{
  switch(expr->variant()) {
  case EXPR_LIST:
    deal_processor_array(s, expr->operand(1));
    if (expr->operand(2))
      deal_processor_array(s, expr->operand(2));
    break;
    
    // zheng.gb
    // scalar processor
  case VAR_REF:
    {
      Processor_def * pdef =new Processor_def;
      pdef->setsymbol(expr->symbol());

      pdef->add_dimention(new SgValueExp(1), new SgValueExp(1));
      if ( in_interface == off )
      {
      	s->controlParent()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
	pdef->set_interface_symbol(NULL);
      }
      else  // the processor is declared in interface
      {
	pdef->set_interface_symbol(s->controlParent()->symbol());
      	s->controlParent()->controlParent()->controlParent()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
      }
      expr->symbol()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
    };
    break;

  case ARRAY_REF:
    {
      Processor_def * pdef =new Processor_def;
      pdef->setsymbol(expr->symbol());
	// parse the number_of_processor 
      intrinsic_functions_in_size(s, expr->operand(1));

      fill_table(pdef, expr->operand(1));
// zheng.gb
      if ( in_interface == off )
      {
      	s->controlParent()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
	pdef->set_interface_symbol(NULL);
      }
      else  // the processor is declared in interface
      {
	pdef->set_interface_symbol(s->controlParent()->symbol());
      	s->controlParent()->controlParent()->controlParent()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
      }
      expr->symbol()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
    break;

  };
}

void analyse_processor(SgStatement * s)
{
  deal_processor_array(s, s->expr(0));
}


void show_processor(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {
    ((Processor_def *) ((s->getAttribute(i, Proc_Def))->getAttributeData()) ) ->show();
  };
  
}

//  add a processor declaration in scope s .
void add_a_processor_declaration(SgStatement *s,Processor_def *pdef)
{
    // The following make up the processor shap declaration;
	 
    SgSymbol * proshp;
    SgSymbol * interface_name = pdef->get_interface_symbol();
    if ( interface_name == NULL )
    	proshp= s->getSymbolInCurrentScope(mixstring, "shp", pdef->getsymbol()->identifier());
    else {	// processor is declared in an interface
	char name[100];
	strcpy(name,interface_name->identifier());
	strcat(name,"_");
	strcat(name,pdef->getsymbol()->identifier());
    	proshp= s->getSymbolInCurrentScope(mixstring, "shp", name);
    }

    pdef->setshp(proshp);

    SgValueExp * dim = new SgValueExp (pdef->getDim());
    SgExpression * range = new SgExpression(RANGE_OP, NULL, dim, NULL);
    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);
    SgExpression * array = new SgExpression(ARRAY_REF, index, NULL, proshp);

    SgType * stype = SgTypeInt();
    SgExpression * arraytype = new SgExpression(TYPE_OP, NULL, NULL, NULL, stype);
  
    SgStatement * declshp = new SgStatement(VAR_DECL, NULL, NULL, array, arraytype, NULL);
  
    s->childList1(0)->insertStmtBefore(*declshp, *s);
  
    // the following is processor grid;
    
    SgSymbol *progrp;
    if (interface_name == NULL)
    	progrp= s->getSymbolInCurrentScope(mixstring, "grp", pdef->getsymbol()->identifier());
    else {
	char name[100];
	strcpy(name,interface_name->identifier());
	strcat(name,"_");
	strcat(name,pdef->getsymbol()->identifier());
    	progrp= s->getSymbolInCurrentScope(mixstring, "grp", name);
    }

    pdef->setgrp(progrp);
    
    SgExpression * var = new SgVarRefExp(progrp);
    SgExpression * expr = new SgExpression(EXPR_LIST, var, NULL, NULL);
  
    SgType * gtype = SgTypeInt();
    SgExpression * exprtype = new SgExpression(TYPE_OP, NULL, NULL, NULL, gtype);

    SgStatement * declgrp =new SgStatement(VAR_DECL, NULL, NULL, expr, exprtype, NULL);
  
    s->childList1(0)->insertStmtBefore(*declgrp, *s);
}

void add_processor_declaration(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {

    Processor_def * pdef = (Processor_def *) s->attributeValue(i, Proc_Def);

    add_a_processor_declaration(s,pdef);

  };
}

//------------------;

// zheng.gb
void get_template_local_size(Template_def *tdef)
{
    SgSymbol * tsymb = tdef->getsymbol();

    if (tsymb->numberOfAttributes(Dist_Def)!=0) {

      Distribute_def * ddef = (Distribute_def *)tsymb->attributeValue(0, Dist_Def);
    
	// dim record the real dimension distributed (except NOP)
      int dim=0;
      for (int j=0; j < tdef->getDim(); j++) {
      
	switch(ddef->distCode(j)) {
	case NOP: {
	  SgExpression *dim_size = &(*tdef->getEnd(j) - *tdef->getStart(j) + *new SgValueExp(1));
	  dim_size = calculateExp(dim_size);
	  tdef->setlocalsize(j, dim_size);
	  break;
	}
	case CYCLIC:
	case BLOCK:
	  { 
	    if (ddef->gettosymbol()->numberOfAttributes(Proc_Def)) {
	      SgExpression *temp;
	      SgExpression *proc = ddef->getProcSize(dim);

	      SgExpression *s,*e;
	      s = tdef->getStart(j) ;
	      e = tdef->getEnd(j) ;

	      if ( s==NULL && e==NULL)
	      {
	      	tdef->setlocalsize(j, new SgValueExp(0));
		continue;
	      }
	      
	      SgExpression *one = new SgValueExp(1);
	      SgExpression *size = &(*e - *s + *one);
	       // get ceiling of size/procnum
	      temp = &( ( *size + *proc - *one->copyPtr())/ (*proc->copyPtr()) );
	      temp = calculateExp(temp);

	      tdef->setlocalsize(j, temp);
	
	    }
	    else
	      printf("get_template_local_size:Not distributed yet\n");
	  };
	  dim++;
	  break;
	default:
	  printf("get_template_local_size:no more code!\n");
	};
      };
    };
}

void caculate_template_local_size(SgStatement * s)
{
  //  show_template(s);

  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    Template_def * tdef = (Template_def *) ((s->attributeValue(i, Temp_Def)));

    get_template_local_size(tdef);
  };
}




